/*
 *    sfall
 *    Copyright (C) 2011  Timeslip
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.h"

#include <stdio.h>
#include "Bugs.h"
#include "Define.h"
#include "FalloutEngine.h"
#include "HookScripts.h"
#include "input.h"
#include "Inventory.h"
#include "LoadGameHook.h"
#include "PartyControl.h"

bool UseScrollWheel = true;

static DWORD InvSizeLimitMode = 0;
static int InvSizeLimit = 0;
static DWORD ReloadWeaponKey = 0;

static DWORD StackEmptyWeapons = 0;

static const int PartyMax = 15;
static int npcCount = 0;
static char iniArmor[MAX_PATH];

struct npcArmor {
 DWORD Pid;
 DWORD Default;
 DWORD Leather;
 DWORD Power;
 DWORD Advanced;
 DWORD Metal;
 DWORD Cured;
 DWORD Combat;
 DWORD Robe;
};

static npcArmor armors[PartyMax];
static char MsgBuf[32];

static void __declspec(naked) critter_max_size() {
 __asm {
  push edx
  mov  edx, InvSizeLimitMode
  dec  edx
  jz   PlayerOnly
  dec  edx
  jz   PartyAndPlayer
  dec  edx
  jz   run                                  //  
fail:
  xor  eax, eax
  dec  eax
  jmp  end
PartyAndPlayer:
  push eax
  call isPartyMember_
  test eax, eax                             // ?
  pop  eax
  jnz  run                                  // 
PlayerOnly:
  cmp  eax, ds:[_obj_dude]
  jne  fail
run:
  mov  edx, STAT_unused
  call stat_level_
end:
  pop  edx
  retn
 }
}

static void __declspec(naked) critter_curr_size() {
 __asm {
  push esi
  push edi
  push edx
  mov  edi, eax
  call item_c_curr_size_
  xchg esi, eax
  mov  eax, [edi+0x20]                      // pobj.fid
  and  eax, 0xF000000
  sar  eax, 0x18
  cmp  eax, ObjType_Critter
  jne  end
  mov  eax, edi
  call inven_right_hand_
  mov  edx, eax
  test eax, eax
  jz   noRight
  test byte ptr [eax+0x27], 2               // iobj.flags4 & Right_Hand_?
  jnz  noRight
  call item_size_
  add  esi, eax
noRight:
  mov  eax, edi
  call inven_left_hand_
  test eax, eax
  jz   skip
  cmp  edx, eax
  je   skip
  test byte ptr [eax+0x27], 1               // iobj.flags4 & Left_Hand_?
  jnz  skip
  call item_size_
  add  esi, eax
skip:
  xchg edi, eax
  call inven_worn_
  test eax, eax
  jz   end
  test byte ptr [eax+0x27], 4               // iobj.flags4 & Worn_?
  jnz  end
  call item_size_
  add  esi, eax
end:
  xchg esi, eax
  pop  edx
  pop  edi
  pop  esi
  retn
 }
}

static void __declspec(naked) stat_level_hook() {
 __asm {
  mov  edx, 0x4AF0D0
  cmp  esi, STAT_max_move_points
  je   end
  cmp  esi, STAT_unused
  jne  skip
  mov  edx, InvSizeLimitMode
  dec  edx
  jz   PlayerOnly
  dec  edx
  jz   PartyAndPlayer
  dec  edx
  jz   run                                  //  
  jmp  skip
PartyAndPlayer:
  push eax
  mov  eax, ebx
  call isPartyMember_
  test eax, eax                             // ?
  pop  eax
  jnz  run                                  // 
PlayerOnly:
  cmp  ebx, edi                             // source == _obj_dude?
  jne  skip
run:
  xchg edx, eax                             // edx = bonus
  mov  eax, InvSizeLimit
  cmp  edx, eax                             // bonus == InvSizeLimit?
  je   skip                                 // 
  push ebx
  push eax
  sub  ecx, edx                             //   
  add  ecx, eax                             //   
  xchg ebx, eax                             // eax = source, ebx = InvSizeLimit
  mov  edx, esi
  call stat_set_bonus_
  pop  eax
  pop  ebx
skip:
  mov  edx, 0x4AF0FC
end:
  jmp  edx
 }
}

static void __declspec(naked) op_set_critter_stat_hook() {
 __asm {
  cmp  dword ptr [esp+4+4], STAT_unused
  jne  end
  pop  ebx                                  //   
  push 0x455D8A
end:
  mov  ebx, 3
  retn
 }
}

static void __declspec(naked) stat_set_base_hook() {
 __asm {
  cmp  ecx, STAT_unused
  jne  end
  pop  eax                                  //   
  push 0x4AF559
end:
  xor  eax, eax
  dec  eax
  retn
 }
}

static char InvenFmt[20]="%d/%d  %d/%d";
static void __declspec(naked) display_stats_hook() {
 __asm {
  mov  ecx, ds:[_curr_stack]
  mov  eax, ds:[_stack][ecx*4]
  push eax
  jecxz itsCritter
  call item_c_max_size_
  pop  edx
  xchg edx, eax                             // eax=source, edx=.  
  push eax
  call item_c_curr_size_
  sub  edx, eax                             // edx=  
  pop  eax
  push edx
  call obj_top_environment_
  push eax
  mov  edx, STAT_carry_amt
  call stat_level_                          // . 
  pop  edx
  xchg edx, eax                             // edx=. , eax=source
  call item_total_weight_
  sub  edx, eax
  mov  eax, HiddenArmor
  call item_weight_
  sub  edx, eax                             // edx= 
  mov  edi, edx
  jmp  print
itsCritter:
  pop  edi
  call critter_max_size
  inc  eax                                  //  CritterInvSizeLimitMode?
  jnz  SizeLimitMode                        // 
  mov  eax, edi                             // source
  mov  edx, STAT_carry_amt
  call stat_level_                          // . 
  push eax
  xchg edi, eax                             // eax=source, edi=. 
  call item_total_weight_
  xchg edx, eax                             // edx= 
  mov  eax, HiddenArmor
  call item_weight_
  add  edx, eax
  sub  edi, edx
print:
  push edx
  mov  edx, [esp+0x9C]
  push edx
  push 0x509EF4                             // "%s %d/%d"
  lea  eax, [esp+0x10]
  push eax
  call sprintf_
  add  esp, 5*4
  movzx ebx, byte ptr ds:[_GreenColor]
  cmp  edi, 0                               //   ?
  jge  noOverloaded                         // 
  mov  bl, ds:[_RedColor]
noOverloaded:
  push 0x472626
  retn
SizeLimitMode:
  dec  eax
  push eax                                  // .
  xchg ecx, eax                             // ecx = .
  mov  eax, edi                             // source
  call critter_curr_size
  xchg edx, eax                             // edx =  
  mov  eax, HiddenArmor
  call item_size_
  add  eax, edx
  push eax                                  //  
  sub  ecx, eax                             // ecx =  
  mov  edx, STAT_carry_amt
  mov  eax, edi                             // source
  call stat_level_
  push eax                                  // .
  xchg ebx, eax                             // ebx = .
  mov  eax, edi                             // source
  call item_total_weight_
  xchg edx, eax                             // edx =  
  mov  eax, HiddenArmor
  call item_weight_
  add  eax, edx
  push eax                                  //  
  sub  ebx, eax                             // ebx =  
  mov  eax, offset InvenFmt
  push eax
  mov  eax, offset MsgBuf
  push eax
  call sprintf_
  add  esp, 6*4
  movzx eax, byte ptr ds:[_GreenColor]
  cmp  ebx, 0                               //   ?
  jl   Red                                  // 
  cmp  ecx, 0                               //   ?
  jge  noRed                                // 
Red:
  mov  al, ds:[_RedColor]
noRed:
  push eax                                  // ColorIndex
  mov  edi, offset MsgBuf
  push edi
  xor  edx, edx
nextChar:
  xor  eax, eax
  mov  al, [edi]
  call ds:[_text_char_width]
  inc  eax
  add  edx, eax
  inc  edi
  cmp  word ptr [edi-1], '  '
  jne  nextChar
  mov  ebx, 150                             // TxtWidth
  mov  ecx, 499                             // ToWidth
  lea  eax, [esi+ebp+75]                    // ToSurface
  sub  eax, edx
  pop  edx                                  // DisplayText
  push 0x472677
  retn
 }
}

static void __declspec(naked) critterIsOverloaded_hook() {
 __asm {
  call item_total_weight_
  cmp  edx, eax                             //   <  ?
  jl   end                                  // 
  mov  eax, ebx                             // source
  call critter_max_size
  inc  eax
  jz   end
  dec  eax
  xchg edx, eax                             // edx = .
  mov  eax, ebx                             // source
  call item_c_curr_size_
end:
  retn
 }
}

static void __declspec(naked) barter_attempt_transaction_peon() {
 __asm {
  call item_total_weight_
  cmp  eax, edx                             //   >  ?
  jg   end                                  // 
  mov  eax, edi                             // source
  call critter_max_size
  inc  eax
  jz   end
  dec  eax
  xchg edx, eax
  mov  eax, edi                             // source
  call critter_curr_size
  sub  edx, eax
  mov  eax, ecx                             // _barterer_table_obj
  call critter_curr_size
end:
  retn
 }
}

static void __declspec(naked) barter_attempt_transaction_barterer() {
 __asm {
  call item_total_weight_
  cmp  eax, edx                             //   >  ?
  jg   end                                  // 
  mov  eax, ebx                             // source
  call critter_max_size
  inc  eax
  jz   end
  dec  eax
  sub  eax, SizeOnBody                      //        
  xchg edx, eax
  mov  eax, ebx                             // source
  call critter_curr_size
  sub  edx, eax
  mov  eax, esi                             // _peon_table_obj
  call critter_curr_size
end:
  retn
 }
}

static void __declspec(naked) loot_container_hook() {
 __asm {
  call item_total_weight_
  cmp  eax, edx                             //   >  ?
  jg   end                                  // 
  mov  eax, [esp+0x134+4]                   // source
  call critter_max_size
  inc  eax
  jz   end
  dec  eax
  xchg edx, eax
  mov  eax, [esp+0x134+4]                   // source
  call critter_curr_size
  sub  edx, eax
  mov  eax, ebp
  call critter_curr_size
end:
  retn
 }
}

static void __declspec(naked) gdControlUpdateInfo_hook() {
 __asm {
  push eax                                  // _dialog_target
  call stat_level_
  xchg ecx, eax                             // ecx = . 
  pop  eax
  push eax
  call critter_max_size
  xchg edi, eax                             // edi = .
  pop  eax
  inc  edi
  jz   end
  dec  edi
  push edi                                  // .
  call item_c_curr_size_
  push eax                                  //  
  push ecx                                  // . 
  push ebx                                  //  
  mov  eax, offset InvenFmt
  push eax
  lea  eax, [esp+0xC+12]
  push eax
  call sprintf_
  add  esp, 6*4
  pop  edx                                  //   
  push 0x449151
  xor  edx, edx
end:
  xchg ecx, eax
  retn
 }
}

static char ItemSizeMsg[32];
static void __declspec(naked) inven_obj_examine_func_hook() {
 __asm {
  xchg esi, eax
  call item_size_
  test eax, eax
  jz   end
  push eax                                  //  
  mov  eax, offset ItemSizeMsg
  push eax
  mov  edx, offset MsgBuf
  push edx
  call sprintf_
  add  esp, 3*4
  xchg edx, eax
  call inven_display_msg_
end:
  xchg esi, eax
  jmp  text_font_
 }
}

static char SuperStimMsg[128];
static void __declspec(naked) protinst_use_item_on_hook() {
 __asm {
  mov  ebx, [ebx+0x64]                      // ebx = item pid
  cmp  ebx, PID_SUPER_STIMPAK
  jne  end
  mov  edx, [edi+0x64]                      // edx = target pid
  shr  edx, 0x18
  cmp  edx, ObjType_Critter
  jne  end
  mov  eax, edi                             // eax = target
  mov  edx, STAT_max_hit_points
  call stat_level_
  cmp  eax, [edi+0x58]                      // max_hp == curr_hp?
  jne  end                                  // 
  pop  eax                                  //   
  mov  eax, offset SuperStimMsg
  push 0x49C478
end:
  xor  edx, edx
  retn
 }
}

static int invenapcost;
static char invenapqpreduction;
void _stdcall SetInvenApCost(int a) {invenapcost = a;}
static void __declspec(naked) handle_inventory_hook() {
 _asm {
  movzx ebx, byte ptr invenapqpreduction
  mul  bl
  mov  edx, invenapcost
  push 0x46E812
  retn
 }
}

// adds check for weapons which require more than 1 ammo for single shot (super cattle prod & mega power fist), or burst
static void __declspec(naked) combat_check_bad_shot_hook() {
 __asm {
  push edx
  push ebx
  call item_w_cur_ammo_
  cmp  esi, ds:[_obj_dude]
  jne  end
  xchg ebx, eax                             // ebx = current ammo
  mov  eax, ecx                             // weapon
  mov  edx, edi                             // hit_mode
  call item_w_anim_weap_
  cmp  eax, ANIM_fire_burst
  je   itsBurst
  cmp  eax, ANIM_fire_continuous
  je   itsBurst
  sub  esp, 4
  xor  eax, eax
  inc  eax
  mov  [esp], eax
  mov  eax, ecx                             // weapon
  mov  edx, esp
  call item_w_compute_ammo_cost_
  mov  eax, [esp]
  add  esp, 4
  jmp  skip
itsBurst:
  mov  eax, ecx                             // weapon
  call item_w_rounds_
skip:
  xchg edx, eax
  xor  eax, eax
  cmp  edx, ebx
  ja   end                                  // this will force "Out of ammo"
  inc  eax                                  // this will force success
end:
  pop  ebx
  pop  edx
  retn
 }
}

static void __declspec(naked) ReloadActiveHand() {
 __asm {
// esi=-1         
  push ebx
  push ecx
  push edx
  mov  eax, ds:[_itemCurrentItem]
  imul ebx, eax, 24
  mov  eax, ds:[_obj_dude]
  xor  ecx, ecx
reloadItem:
  push eax
  mov  edx, ds:[_itemButtonItems][ebx]
  call item_w_try_reload_
  test eax, eax
  pop  eax
  jnz  endReloadItem
  inc  ecx
  jmp  reloadItem
endReloadItem:
  cmp  dword ptr ds:[_itemButtonItems + 0x10][ebx], 5// mode
  jne  skip_toggle_item_state
  call intface_toggle_item_state_
skip_toggle_item_state:
  test ecx, ecx
  jnz  useActiveHand
  xchg esi, ebx
useActiveHand:
  push ebx
  xor  esi, esi
  dec  esi
  mov  ebx, esi
  xor  eax, eax
  mov  edx, ebx
  call intface_update_items_
  pop  eax
  cmp  eax, esi
  je   end
  mov  ebx, 2
  xor  ecx, ecx
  mov  edx, ds:[_itemButtonItems][eax]
  push 0x460B85
  retn
end:
  pop  edx
  pop  ecx
  pop  ebx
  retn
 }
}

static void __declspec(naked) ReloadWeaponHotKey() {
 __asm {
  call gmouse_is_scrolling_
  test eax, eax
  jnz  end
  pushad
  xchg ebx, eax
  push ReloadWeaponKey
  call KeyDown
  test eax, eax
  jnz  ourKey
  popad
  retn
ourKey:
  cmp  ds:[_intfaceEnabled], ebx
  je   endReload
  xor  esi, esi
  dec  esi
  cmp  ds:[_interfaceWindow], esi
  je   endReload
  mov  edx, ds:[_itemCurrentItem]
  imul eax, edx, 24
  cmp  ds:[_itemButtonItems + 0x5][eax], bl // itsWeapon
  jne  itsWeapon                            // 
  call intface_use_item_
  jmp  endReload
itsWeapon:
  test byte ptr ds:[_combat_state], 1       //  ?
  jnz  inCombat                             // 
  call ReloadActiveHand
  jmp  endReload
inCombat:
//  xor  ebx, ebx                             // is_secondary
  add  edx, hit_left_weapon_reload          // edx = 6/7 -    / 
  mov  eax, ds:[_obj_dude]
  push eax
  call item_mp_cost_
  xchg ecx, eax
  pop  eax                                  // _obj_dude
  mov  edx, [eax+0x40]                      // movePoints
  cmp  ecx, edx
  jg   endReload
  push eax
  call ReloadActiveHand
  test eax, eax
  pop  eax                                  // _obj_dude
  jnz  endReload
  sub  edx, ecx
  mov  [eax+0x40], edx                      // movePoints
  xchg edx, eax
  mov  edx, ds:[_combat_free_move]
  call intface_update_move_points_
endReload:
  popad
  inc  eax
end:
  retn
 }
}

static void __declspec(naked) AutoReloadWeapon() {
 __asm {
  call scr_exec_map_update_scripts_
  pushad
  cmp  dword ptr ds:[_game_user_wants_to_quit], 0
  jnz  end
  mov  eax, ds:[_obj_dude]
  call critter_is_dead_                     //    
  test eax, eax
  jnz  end
  cmp  ds:[_intfaceEnabled], eax
  je   end
  xor  esi, esi
  dec  esi
  cmp  ds:[_interfaceWindow], esi
  je   end
  inc  eax
  mov  ebx, ds:[_itemCurrentItem]
  push ebx
  sub  eax, ebx
  mov  ds:[_itemCurrentItem], eax           //   
  imul ebx, eax, 24
  mov  eax, ds:[_obj_dude]
  xor  ecx, ecx
reloadOffhand:
  push eax
  mov  edx, ds:[_itemButtonItems][ebx]
  call item_w_try_reload_
  test eax, eax
  pop  eax
  jnz  endReloadOffhand
  inc  ecx
  jmp  reloadOffhand
endReloadOffhand:
  cmp  dword ptr ds:[_itemButtonItems + 0x10][ebx], 5// mode
  jne  skip_toggle_item_state
  call intface_toggle_item_state_
skip_toggle_item_state:
  test ecx, ecx
  jnz  useOffhand
  xchg ebx, esi
useOffhand:
  xchg esi, ebx                             // esi=-1       
  pop  eax
  mov  ds:[_itemCurrentItem], eax           //   
  call ReloadActiveHand
end:
  popad
  retn
 }
}

static int pobj;
static int _stdcall ChangeArmorFid(DWORD* item, DWORD* npc) {
 int itm_pid = 0, npc_pid = PID_Player;
 if (item) itm_pid = item[0x64/4]&0xffffff;
 if (npc) npc_pid = npc[0x64/4];
 pobj = 0;
 for (int k = 0; k < npcCount; k++) {
  if (armors[k].Pid == npc_pid) {
   pobj = armors[k].Default;
   switch (itm_pid) {
    case PID_LEATHER_ARMOR: case PID_LEATHER_ARMOR_MK_II:
     if (armors[k].Leather != 0) pobj = armors[k].Leather;
     break;
    case PID_POWERED_ARMOR: case PID_HARDENED_POWER_ARMOR:
     if (armors[k].Power != 0) pobj = armors[k].Power;
     break;
    case PID_ADVANCED_POWER_ARMOR: case PID_ADVANCED_POWER_ARMOR_MK2:
     if (armors[k].Advanced != 0) pobj = armors[k].Advanced;
     break;
    case PID_METAL_ARMOR: case PID_TESLA_ARMOR: case PID_METAL_ARMOR_MK_II:
     if (armors[k].Metal != 0) pobj = armors[k].Metal;
     break;
    case PID_LEATHER_JACKET: case PID_CURED_LEATHER_ARMOR:
     if (armors[k].Cured != 0) pobj = armors[k].Cured;
     break;
    case PID_COMBAT_ARMOR: case PID_BROTHERHOOD_COMBAT_ARMOR: case PID_COMBAT_ARMOR_MK_II:
     if (armors[k].Combat != 0) pobj = armors[k].Combat;
     break;
    case PID_PURPLE_ROBE: case PID_BRIDGEKEEPERS_ROBE:
     if (armors[k].Robe != 0) pobj = armors[k].Robe;
     break;
   }
   __asm {
    mov  ebx, npc
    mov  eax, ebx
    call inven_right_hand_
    push eax                                //    ,    
    test eax, eax
    jz   noWeapon                           //    
    mov  eax, ebx
    xor  edx, edx
    inc  edx                                //   (INVEN_TYPE_RIGHT_HAND)
    call inven_unwield_                     // 
noWeapon:
    push pobj                               // pobj = fid 
    mov  edx, offset pobj
    mov  eax, npc_pid
    call proto_ptr_
    mov  edx, pobj
    mov  eax, [edx+0x8]
    mov  pobj, eax                          // pobj = fid  pro-
    pop  edx
    test edx, edx                           //    ?
    jnz  haveFid                            // 
    mov  edx, eax                           // ,  fid  pro-
haveFid:
    and  edx, 0x0FFF                        // edx =   LST-
    mov  ebx, [ebx+0x20]                    // fid
    mov  eax, ebx
    and  eax, 0x70000000
    sar  eax, 0x1C
    push eax                                // ID3 (Direction code)
    and  ebx, 0x0FF0000
    sar  ebx, 0x10                          // ebx = ID2 (Animation code)
    xor  ecx, ecx                           // ecx = ID1 (Weapon code = None)
    mov  eax, ObjType_Critter               // eax =  
    call art_id_
    mov  edx, eax
    call art_exists_
    test eax, eax
    jnz  canUse                             //    
    mov  edx, pobj                          //    pro-
canUse:
    mov  pobj, edx                          //  fid  
    xor  ebx, ebx
    mov  eax, npc
    mov  ecx, eax
    call obj_change_fid_                    //  
    pop  edx                                //    ,    
    test edx, edx                           //    ?
    jz   end                                // 
    xor  ebx, ebx
    inc  ebx                                //   (INVEN_TYPE_RIGHT_HAND)
    mov  eax, ecx                           // *npc
    call inven_wield_                       //  
    mov  edx, [ecx+0x20]                    // fid
    mov  pobj, edx                          //  fid     correctFidForRemovedItem_
end:
   }
   break;
  }
 }
 return pobj;                               //  fid
}

static void __declspec(naked) correctFidForRemovedItem_hook() {
 __asm {
  call adjust_ac_
nextArmor:
  mov  eax, esi
  call inven_worn_
  test eax, eax
  jz   noArmor
  and  byte ptr [eax+0x27], 0xFB            //    
  jmp  nextArmor
noArmor:
  cmp  edi, -1                              //  ?
  jnz  end                                  // 
  cmp  npcCount, eax
  je   end
  push esi                                  //   
  push eax                                  //   
  call ChangeArmorFid
  test eax, eax
  jz   end
  mov  edi, eax
end:
  retn
 }
}

static void __declspec(naked) ControlWeapon_hook() {
 __asm {
  push eax
  call inven_right_hand_
  xchg ecx, eax
  pop  eax
  call inven_unwield_
  test eax, eax                             //  ? (hs_invenwield)
  jnz  skip                                 // 
  jecxz end                                 //      
skip:
  pop  eax                                  //   
  push 0x44952E
end:
  retn
 }
}

static void __declspec(naked) ControlArmor_hook() {
 __asm {
  mov  ecx, eax                             // _dialog_target
  call inven_worn_
  test eax, eax
  jnz  haveArmor
  xchg ecx, eax                             // _dialog_target
  jmp  ai_search_inven_armor_
haveArmor:
  mov  ebx, 0x4000000                       // Worn
  xchg edx, eax                             // edx =    
  xchg ecx, eax                             // eax = _dialog_target
  jmp  correctFidForRemovedItem_
 }
}

static void __declspec(naked) fontHeight() {
 __asm {
  push ebx
  mov  ebx, ds:[_curr_font_num]
  mov  eax, 101
  call text_font_
  call ds:[_text_height]
  xchg ebx, eax
  call text_font_
  xchg ebx, eax
  pop  ebx
  retn
 }
}

static void __declspec(naked) printFreeMaxWeightSize() {
 __asm {
// ebx = source, ecx = ToWidth, edi = posOffset, esi = extraWeight, ebp = extraSize
  mov  eax, ds:[_curr_font_num]
  push eax
  mov  eax, 101
  call text_font_
  mov  eax, ds:[_i_wid]
  call win_get_buf_                         // eax=ToSurface
  add  edi, eax                             // ToSurface+posOffset (Ypos*ToWidth+Xpos)
  mov  eax, [ebx+0x20]
  and  eax, 0xF000000
  sar  eax, 0x18
  test eax, eax                             //  ObjType_Item?
  jz   itsItem                              // 
  dec  eax                                  //  ObjType_Critter?
  jnz  noWeight                             // 
  mov  eax, ebx
  mov  edx, STAT_carry_amt
  call stat_level_
  xchg ebx, eax                             // ebx = .  , eax = source
  sub  ebx, esi                             // ebx = .   - extraWeight
  push eax
  call critter_max_size
  xchg edx, eax                             // edx = .  
  pop  eax
  inc  edx
  jnz  WeightAndSize
  call item_total_weight_                   // eax =   
  jmp  printFree
WeightAndSize:
  dec  edx
  sub  edx, ebp                             // edx = .   - extraSize
  push eax
inContainer:
  call critter_curr_size
  sub  edx, eax                             // edx =  
  pop  eax
  push edx
  call item_total_weight_
  sub  ebx, eax                             // eax =  
  push ebx
  push 0x503180                             // '%d/%d'
  mov  esi, offset MsgBuf
  push esi
  call sprintf_
  add  esp, 4*4
  movzx eax, byte ptr ds:[_GreenColor]
  cmp  ebx, 0                               //   ?
  jl   Red                                  // 
  cmp  edx, 0                               //   ?
  jge  noRed                                // 
Red:
  mov  al, ds:[_RedColor]
noRed:
  push eax
  push esi
  xor  edx, edx
nextChar:
  movzx eax, byte ptr [esi]
  call ds:[_text_char_width]
  inc  eax
  add  edx, eax
  inc  esi
  cmp  byte ptr [esi-1], '/'
  jne  nextChar
  sub  edi, edx
  pop  edx                                  // DisplayText
  jmp  print
itsItem:
  mov  eax, ebx
  call item_get_type_
  dec  eax                                  // item_type_container?
  jnz  noWeight                             // 
  mov  eax, ebx
  call obj_top_environment_
  test eax, eax                             //  ?
  jz   noOwner                              // 
  mov  edx, [eax+0x20]
  and  edx, 0xF000000
  sar  edx, 0x18
  cmp  edx, ObjType_Critter                 //  ?
  jne  noOwner                              // 
  push eax
  mov  edx, STAT_carry_amt
  call stat_level_
  xchg ebx, eax                             // ebx = .    , eax = source
  sub  ebx, esi                             // ebx = .   - extraWeight
  push eax
  call item_c_max_size_
  xchg edx, eax                             // edx = .   
  pop  eax
  jmp  inContainer
noOwner:
  mov  eax, ebx
  call item_c_max_size_                     // eax = .  
  xchg ebx, eax                             // ebx = .  , eax = source
  call item_c_curr_size_
printFree:
  sub  ebx, eax                             // ebx =  /
  push ebx
  push 0x503190                             // '%d'
  mov  edx, offset MsgBuf
  push edx
  call sprintf_
  add  esp, 3*4
  movzx eax, byte ptr ds:[_GreenColor]
  cmp  ebx, 0                               //   ?
  jge  skipColor                            // 
  mov  al, ds:[_RedColor]
skipColor:
  push eax
  mov  eax, edx
  call ds:[_text_width]
  shr  eax, 1
  inc  eax
  sub  edi, eax
print:
  xchg edi, eax                             // ToSurface+posOffset (Ypos*ToWidth+Xpos)
  mov  ebx, 64                              // TxtWidth
  call ds:[_text_to_buf]
noWeight:
  pop  eax
  jmp  text_font_
 }
}

static void __declspec(naked) display_inventory_hook() {
 __asm {
  call fontHeight
  inc  eax
  add  [esp+0x8], eax                       // height = height + text_height + 1
  call buf_to_buf_
  add  esp, 0x18
  mov  eax, [esp+0x4]
  call art_ptr_unlock_
  pushad
  mov  eax, ds:[_curr_stack]
  mov  ebx, ds:[_stack][eax*4]
  mov  ecx, 537
  mov  edi, 325*537+180+32                  // Xpos=180, Ypos=325, max text width/2=32
  xor  esi, esi
  xor  ebp, ebp
  call printFreeMaxWeightSize
  popad
  push 0x4700C5
  retn
 }
}

static void __declspec(naked) display_target_inventory_hook() {
 __asm {
  call fontHeight
  inc  eax
  add  [esp+0x8], eax                       // height = height + text_height + 1
  call buf_to_buf_
  add  esp, 0x18
  mov  eax, [esp]
  call art_ptr_unlock_
  pushad
  mov  eax, ds:[_target_curr_stack]
  mov  ebx, ds:[_target_stack][eax*4]
  mov  ecx, 537
  mov  edi, 325*537+301+32                  // Xpos=301, Ypos=325, max text width/2=32
  mov  esi, WeightOnBody                    //        
  mov  ebp, SizeOnBody                      //        
  call printFreeMaxWeightSize
  popad
  push 0x470468
  retn
 }
}

static void __declspec(naked) display_table_inventories_hook() {
 __asm {
  call win_get_buf_
  mov  edi, ds:[_btable]
  mov  [esp+0x8C+4], edi
  mov  edi, ds:[_ptable]
  mov  [esp+0x94+4], edi
  retn
 }
}

static void __declspec(naked) display_table_inventories_hook1() {
 __asm {
  add  dword ptr [esp+8], 20
  sub  dword ptr [esp+16], 20*480
  call buf_to_buf_
  add  esp, 0x18
  pushad
  mov  eax, ds:[_btable]
  push eax
  call item_total_weight_                   // eax =      
  xchg esi, eax
  pop  eax
  call critter_curr_size                    // eax =      
  xchg ebp, eax
  mov  eax, ds:[_curr_stack]
  mov  ebx, ds:[_stack][eax*4]
  mov  ecx, 480
  mov  edi, 10*480+169+32                   // Xpos=169, Ypos=10, max text width/2=32
  call printFreeMaxWeightSize
  popad
  push 0x47548C
  retn
 }
}

//   
static void __declspec(naked) display_table_inventories_hook2() {
 __asm {
  mov  dword ptr [edx+4], 4                 // WinRect.y_start = 4
  jmp  win_draw_rect_
 }
}

static void __declspec(naked) display_table_inventories_hook3() {
 __asm {
  add  dword ptr [esp+8], 20
  sub  dword ptr [esp+16], 20*480
  call buf_to_buf_
  add  esp, 0x18
#ifndef TRACE
  cmp  dword ptr ds:[_dialog_target_is_party], 0
  je   end                                  //  
#endif
  pushad
  mov  eax, ds:[_ptable]
  push eax
  call item_total_weight_                   // eax =      
  xchg esi, eax
  add  esi, WeightOnBody                    //        
  pop  eax
  call critter_curr_size
  add  eax, SizeOnBody                      //        
  xchg ebp, eax
  mov  eax, ds:[_target_curr_stack]
  mov  ebx, ds:[_target_stack][eax*4]
  mov  ecx, 480
  mov  edi, 10*480+254+32                   // Xpos=254, Ypos=10, max text width/2=32
  call printFreeMaxWeightSize
  popad
#ifndef TRACE
end:
#endif
  push 0x475612
  retn
 }
}

static void __declspec(naked) barter_inventory_hook() {
 __asm {
  call container_exit_
  xor  ecx, ecx
  dec  ecx
  mov  ebx, [esp+0x30+0x4]                  // target
  mov  eax, edi                             // owner
  mov  edx, ebp                             // source
  jmp  display_table_inventories_
 }
}

static void __declspec(naked) gdProcess_hook() {
 __asm {
  mov  eax, ds:[_dialog_target]
  cmp  dword ptr [eax+0x64], PID_Marcus
  je   noArmor
  mov  edx, eax
  call ai_search_inven_armor_
  test eax, eax
  jz   noArmor
  xor  ebx, ebx
  xchg edx, eax                             // edx=   (), eax=_dialog_target
  call inven_wield_
noArmor:
  jmp  gdialog_barter_cleanup_tables_
 }
}

static void __declspec(naked) invenWieldFunc_hook() {
 __asm {
  call adjust_ac_
  push esi                                  //   
  push edi                                  //    ()
  call ChangeArmorFid
  retn
 }
}

static void __declspec(naked) SetDefaultAmmo() {
 __asm {
  push eax
  push ebx
  push edx
  xchg edx, eax
  mov  ebx, eax
  call item_get_type_
  cmp  eax, item_type_weapon
  jne  end                                  // 
  cmp  dword ptr [ebx+0x3C], 0              //    ?
  jne  end                                  // 
  sub  esp, 4
  mov  edx, esp
  mov  eax, [ebx+0x64]                      // eax = pid 
  call proto_ptr_
  mov  edx, [esp]
  mov  eax, [edx+0x5C]                      // eax =     
  mov  [ebx+0x40], eax                      //   
  add  esp, 4
end:
  pop  edx
  pop  ebx
  pop  eax
  retn
 }
}

static void __declspec(naked) inven_action_cursor_hook() {
 __asm {
  mov  edx, [esp+0x1C]
  call SetDefaultAmmo
  cmp  dword ptr [esp+0x18], 0
  push 0x4736CB
  retn
 }
}

static void __declspec(naked) inven_pickup_hook2() {
 __asm {
  mov  eax, ds:[_i_wid]
  call GNW_find_
  mov  ebx, [eax+0x8+0x0]                   // ebx = _i_wid.rect.x
  mov  ecx, [eax+0x8+0x4]                   // ecx = _i_wid.rect.y
  mov  eax, 176
  add  eax, ebx                             // x_start
  add  ebx, 176+60                          // x_end
  mov  edx, 37
  add  edx, ecx                             // y_start
  add  ecx, 37+100                          // y_end
  call mouse_click_in_
  test eax, eax
  jz   end
  mov  edx, ds:[_curr_stack]
  test edx, edx
  jnz  end
  cmp  edi, 1006                            // ?
  jae  skip                                 // 
  mov  edx, [esp+0x18+0x4]                  // item
  mov  eax, edx
  call item_get_type_
  cmp  eax, item_type_drug
  jne  skip
  mov  eax, ds:[_stack]
  push eax
  push edx
  call useobjon_item_d_take_drug_           // item_d_take_drug_
  pop  edx
  pop  ebx
  dec  eax
  jnz  notUsed
  inc  eax
  xchg ebx, eax
  push edx
  push eax
  call item_remove_mult_
  inc  eax
  pop  eax
  mov  ebx, [eax+0x28]
  mov  edx, [eax+0x4]
  pop  eax
  jz   notUsed
  xor  ecx, ecx
  push eax
  call obj_connect_
  pop  eax
  call obj_destroy_
notUsed:
  xor  eax, eax
  inc  eax
  call intface_update_hit_points_
skip:
  xor  eax, eax
end:
  retn
 }
}

static void __declspec(naked) make_drop_button() {
 __asm {
  mov  ebx, [esp+0x28+0x4]
  mov  eax, [ebx+0x20]
  and  eax, 0xF000000
  sar  eax, 0x18
  test eax, eax                             //  ObjType_Item?
  jnz  skip                                 // 
  xchg ebx, eax
  call item_get_type_
  dec  eax                                  // item_type_container?
  jz   goodTarget                           // 
  jmp  noButton
skip:
  cmp  eax, ObjType_Critter
  jne  noButton                             // 
  xchg ebx, eax
  call critter_body_type_
  test eax, eax                             //  Body_Type_Biped?
  jnz  noButton                             // 
goodTarget:
  xor  ecx, ecx                             // ID1
  xor  ebx, ebx                             // ID2
  push ebx                                  // ID3
  mov  edx, 255                             // Index (DROPN.FRM (Action menu drop normal))
  mov  eax, ObjType_Intrface                // ObjType
  call art_id_
  xor  edx, edx
  mov  ecx, 0x59E7E4
  xor  ebx, ebx
  call art_ptr_lock_data_
  test eax, eax
  jz   noButton
  xchg esi, eax
  xor  ecx, ecx                             // ID1
  xor  ebx, ebx                             // ID2
  push ebx                                  // ID3
  mov  edx, 254                             // Index (DROPH.FRM (Action menu drop highlighted))
  mov  eax, ObjType_Intrface                // ObjType
  call art_id_
  xor  edx, edx
  mov  ecx, 0x59E7E8
  xor  ebx, ebx
  call art_ptr_lock_data_
  test eax, eax
  jz   noButton
  push ebp                                  // ButType
  push ebp
  push eax                                  // PicDown
  push esi                                  // PicUp
  dec  ebp
  push ebp                                  // ButtUp
  mov  edx, 68                              // Xpos
  push edx                                  // ButtDown
  push ebp                                  // HovOff
  push ebp                                  // HovOn
  inc  ebp
  mov  ecx, 40                              // Width
  push ecx                                  // Height
  mov  ebx, 204                             // Ypos
  mov  eax, ds:[_i_wid]                     // GNWID
  call win_register_button_
noButton:
  mov  edx, 436
  retn
 }
}

static char OverloadedDrop[48];
static void __declspec(naked) drop_all() {
 __asm {
  cmp  esi, 'a'
  jne  skip
  push 0x4740DB
  retn
skip:
  cmp  esi, 'D'
  je   dropKey
  cmp  esi, 'd'
  je   dropKey
  push 0x4741B2
  retn
dropKey:
  pushad
  cmp  dword ptr ds:[_gIsSteal], 0
  jne  end
  mov  esi, ebp
  mov  ecx, [esp+0x134+0x20]
  mov  eax, ecx
  call critter_body_type_
  test eax, eax                             //  Body_Type_Biped?
  jnz  end                                  // 
  mov  eax, [esi+0x20]
  and  eax, 0xF000000
  sar  eax, 0x18
  test eax, eax                             //  ObjType_Item?
  jz   itsItem                              // 
  cmp  eax, ObjType_Critter
  jne  end                                  // 
  mov  eax, esi
  call critter_body_type_
  test eax, eax                             //  Body_Type_Biped?
  jnz  end                                  // 
itsCritter:
  mov  edx, STAT_carry_amt
  mov  eax, esi
  call stat_level_
  xchg edx, eax                             // edx = .   
  sub  edx, WeightOnBody                    //        
  mov  eax, esi
  call item_total_weight_                   // eax =    
  sub  edx, eax
  mov  eax, ecx
  call item_total_weight_
  xchg edi, eax                             // edi =    
  mov  eax, ds:[_i_rhand]
  call item_weight_
  sub  edi, eax
  mov  eax, ds:[_i_lhand]
  call item_weight_
  sub  edi, eax
  mov  eax, ds:[_i_worn]
  call item_weight_
  sub  edi, eax
  xchg edi, eax                             // eax =    
  cmp  eax, edx
  jg   compareSizeWeight
  mov  eax, esi
  call critter_max_size
  inc  eax
  jz   cont
  dec  eax
  xchg edx, eax
  sub  edx, SizeOnBody                      //        
  mov  eax, esi                             // source
  call critter_curr_size
  sub  edx, eax
  mov  eax, ecx
  call item_c_curr_size_
  jmp  compareSizeWeight
itsItem:
  mov  eax, esi
  call item_get_type_
  dec  eax                                  //  /?
  jnz  end                                  // 
  mov  eax, esi
  call obj_top_environment_
  test eax, eax                             //  ?
  jz   noOwner                              // 
  xchg esi, eax                             // esi =  
  mov  eax, [esi+0x20]
  and  eax, 0xF000000
  sar  eax, 0x18
  cmp  eax, ObjType_Critter                 //  ?
  je   itsCritter                           // 
noOwner:
  mov  eax, esi
  call item_c_max_size_
  xchg edx, eax
  mov  eax, esi
  call item_c_curr_size_
  sub  edx, eax
  mov  eax, ecx
  call item_c_curr_size_
compareSizeWeight:
  cmp  eax, edx
  jg   cantDrop
cont:
  call gsound_red_butt_press_
  mov  edx, ebp
  mov  eax, ecx
  call item_move_all_
  xor  eax, eax
  mov  ds:[_stack_offset], eax
  mov  ds:[_target_stack_offset], eax
  inc  eax
  mov  ds:[_curr_stack], eax
  inc  eax
  push eax                                  // mode
  xchg ecx, eax                             // ecx = mode
  mov  eax, ds:[_target_curr_stack]
  mov  eax, ds:[_target_stack_offset][eax*4]// inventory_offset
  xor  edx, edx
  dec  edx                                  // visible_offset
  mov  ebx, ds:[_target_pud]                // target_inventory
  call display_target_inventory_
  pop  edx                                  // 2 (mode)
  mov  eax, 2500
  call container_exit_
  jmp  end
cantDrop:
  call gsound_red_butt_press_
  xor  eax, eax
  push eax
  mov  al, ds:[0x6AB718]                    // color
  push eax
  xor  ebx, ebx
  push ebx
  push eax
  mov  ecx, 169
  push 117
  xor  edx, edx
  lea  eax, OverloadedDrop
  call dialog_out_
end:
  popad
  push 0x474435
  retn
 }
}

static void __declspec(naked) loot_container_hook1() {
 __asm {
  xor  eax, eax
  mov  ds:[_stack_offset], eax
  mov  ds:[_target_stack_offset], eax
  inc  eax
  mov  ds:[_target_curr_stack], eax
  mov  eax, 2501
  mov  edx, ecx                             // edx = mode
  jmp  container_exit_
 }
}

static void __declspec(naked) loot_container_hook2() {
 __asm {
  cmp  esi, 0x150                           // source_down
  je   scroll
  cmp  esi, 0x148                           // source_up
  jne  end
scroll:
  push edx
  push ecx
  push ebx
  mov  eax, ds:[_i_wid]
  call GNW_find_
  mov  ebx, [eax+0x8+0x0]                   // ebx = _i_wid.rect.x
  mov  ecx, [eax+0x8+0x4]                   // ecx = _i_wid.rect.y
  mov  eax, 297
  add  eax, ebx                             // x_start
  add  ebx, 297+64                          // x_end
  mov  edx, 37
  add  edx, ecx                             // y_start
  add  ecx, 37+6*48                         // y_end
  call mouse_click_in_
  pop  ebx
  pop  ecx
  pop  edx
  test eax, eax
  jz   end
  cmp  esi, 0x150                           // source_down
  je   targetDown
  mov  esi, 0x18D                           // target_up
  jmp  end
targetDown:
  mov  esi, 0x191                           // target_down
end:
  mov  eax, ds:[_curr_stack]
  retn
 }
}

static void __declspec(naked) barter_inventory_hook1() {
 __asm {
  push edx
  push ecx
  push ebx
  xchg esi, eax
  cmp  esi, 0x150                           // source_down
  je   scroll
  cmp  esi, 0x148                           // source_up
  jne  end
scroll:
  mov  eax, ds:[_i_wid]
  call GNW_find_
  mov  ebx, [eax+0x8+0x0]                   // ebx = _i_wid.rect.x
  mov  ecx, [eax+0x8+0x4]                   // ecx = _i_wid.rect.y
  push ebx
  push ecx
  mov  eax, 395
  add  eax, ebx                             // x_start
  add  ebx, 395+64                          // x_end
  mov  edx, 35
  add  edx, ecx                             // y_start
  add  ecx, 35+3*48                         // y_end
  call mouse_click_in_
  pop  ecx
  pop  ebx
  test eax, eax
  jz   notTargetScroll
  cmp  esi, 0x150                           // source_down
  je   targetDown
  mov  esi, 0x18D                           // target_up
  jmp  end
targetDown:
  mov  esi, 0x191                           // target_down
  jmp  end
notTargetScroll:
  push ebx
  push ecx
  mov  eax, 250
  add  eax, ebx                             // x_start
  add  ebx, 250+64                          // x_end
  mov  edx, 20
  add  edx, ecx                             // y_start
  add  ecx, 20+3*48                         // y_end
  call mouse_click_in_
  pop  ecx
  pop  ebx
  test eax, eax
  jz   notTargetBarter
  cmp  esi, 0x150                           // source_down
  je   barterTargetDown
  mov  esi, 0x184                           // target_barter_up
  jmp  end
barterTargetDown:
  mov  esi, 0x176                           // target_barter_down
  jmp  end
notTargetBarter:
  mov  eax, 165
  add  eax, ebx                             // x_start
  add  ebx, 165+64                          // x_end
  mov  edx, 20
  add  edx, ecx                             // y_start
  add  ecx, 20+3*48                         // y_end
  call mouse_click_in_
  test eax, eax
  jz   end
  cmp  esi, 0x150                           // source_down
  je   barterSourceDown
  mov  esi, 0x149                           // source_barter_up
  jmp  end
barterSourceDown:
  mov  esi, 0x151                           // source_barter_down
end:
  pop  ebx
  pop  ecx
  pop  edx
  mov  eax, esi
  cmp  eax, 0x11
  retn
 }
}

static void __declspec(naked) handle_inventory_hook1() {
 __asm {
  call container_exit_
  jmp  display_stats_
 }
}

static void __declspec(naked) itsLootBarterTarget() {
 __asm {
  push edx
  xor  edx, edx
  test eax, eax
  jz   end
  test InLoop, 0x30000                      // INTFACELOOT + BARTER
  jz   end
  cmp  eax, ds:[_target_stack]
  jne  end
  inc  edx
end:
  xchg edx, eax
  pop  edx
  retn
 }
}

void __declspec(naked) item_add_check() {
 __asm {
  push edi
  push esi
  push edx
  push ecx
  push ebx
  mov  esi, eax
  mov  edi, edx
  cmp  ebx, 1
  jl   minus1
  mov  eax, [esi+0x20]                      // eax = pobj.fid
  and  eax, 0xF000000
  sar  eax, 0x18
  cmp  eax, ObjType_Critter
  jne  notCritter
  mov  eax, esi
  call critter_body_type_
  test eax, eax                             // Body_Type_Biped?
  jz   itsCritter
  mov  eax, -5
  jmp  end
itsCritter:
  mov  eax, esi
  call critter_max_size
  inc  eax
  jz   checkWeight
  dec  eax
  xchg ecx, eax                             // ecx = .
  xchg edx, eax                             // eax = item
  call item_size_
  imul eax, ebx                             // eax =   * 
  xchg edx, eax                             // edx =   * 
  mov  eax, esi
  call critter_curr_size
  add  edx, eax
  mov  eax, esi
  call itsLootBarterTarget
  test eax, eax
  jz   noExtraSizeTarget
  dec  eax
  add  edx, SizeOnBody                      //        
noExtraSizeTarget:
  cmp  edx, ecx                             //   <= .?
  jle  checkWeight                          // 
  mov  eax, -4
  jmp  end
checkWeight:
  xchg ecx, eax
  mov  eax, edi
  call item_weight_
  imul eax, ebx                             // eax =   * 
  add  ecx, eax
  mov  eax, esi
  call item_total_weight_
  add  ecx, eax
  mov  eax, esi
  call itsLootBarterTarget
  test eax, eax
  jz   noExtraWeightTarget
  add  ecx, WeightOnBody
noExtraWeightTarget:
  mov  edx, STAT_carry_amt
  mov  eax, esi
  call stat_level_
  cmp  ecx, eax                             //   <= .?
  jle  force                                // 
  mov  eax, -6
  jmp  end
notCritter:
  test eax, eax                             // ObjType_Item?
  jnz  force                                //  (  -   minus1  force)
  mov  eax, esi
  call item_get_type_
  cmp  eax, item_type_container
  jne  notContainer
  xchg edx, eax                             // eax = item
  call item_size_
  xchg edx, eax                             // edx =  
  imul edx, ebx                             // edx =   * 
  mov  eax, esi
  call item_c_curr_size_
  add  edx, eax
  mov  eax, esi
  call item_c_max_size_
  cmp  edx, eax                             //   <= .?
  jle  ownerContainer                       // 
  xor  eax, eax
  dec  eax
  dec  eax                                  // eax = -2 (   )
  jmp  end
ownerContainer:
  mov  eax, esi
  call obj_top_environment_
  test eax, eax                             //   ?
  jz   force                                // 
  xchg esi, eax                             // esi =  
  mov  eax, [esi+0x20]                      // eax = pobj.fid
  and  eax, 0xF000000
  sar  eax, 0x18
  cmp  eax, ObjType_Critter
  jne  force
  dec  eax
  cmp  esi, ds:[_stack]                     //   ?
  jne  checkWeight                          // 
  xchg ecx, eax
  test InLoop, 0x1000                       // INVENTORY
  jz   notFound
  cmp  ds:[_curr_stack], ecx                //           ?
  jne  notFound                             // 
  cmp  edi, ds:[_i_rhand]
  je   found
  cmp  edi, ds:[_i_lhand]
  je   found
  cmp  edi, ds:[_i_worn]
  jne  notFound
found:
  mov  eax, edi                             //  
  call item_weight_
  imul eax, ebx                             // eax =   * 
  sub  ecx, eax                             //  
notFound:
  mov  eax, HiddenArmor
  call item_weight_
  add  eax, ecx
  jmp  checkWeight
notContainer:
  cmp  eax, item_type_misc_item
  jne  minus1
  mov  eax, esi
  call item_m_cell_pid_
  cmp  eax, [edi+0x64]
  je   force
minus1:
  xor  eax, eax
  dec  eax
  jmp  end
force:
  xor  eax, eax
end:
  pop  ebx
  pop  ecx
  pop  edx
  pop  esi
  pop  edi
  retn
 }
}

static void __declspec(naked) item_add_mult() {
 __asm {
  mov  ecx, eax
  call item_add_check
  test eax, eax                             //  ?
  jnz  end                                  // 
  xchg ecx, eax
  cmp  StackEmptyWeapons, ecx
  je   skip
  call SetDefaultAmmo
skip:
  call item_add_force_
end:
  pop  ecx
  retn
 }
}

static void __declspec(naked) move_table_source() {
 __asm {
  xchg ecx, eax                             // ecx = source, eax = count
  xchg ebx, eax                             // ebx = count, eax = item
  xchg edx, eax                             // edx = item, eax = target
  mov  ebp, eax                             // ebp = target
  call item_add_check
  cmp  eax, -2                              //     /?
  je   end                                  // 
  xchg ebp, eax                             // eax = target
  xchg edx, eax                             // edx = target, eax = item
  xchg ebx, eax                             // ebx = item, eax = count
  xchg ecx, eax                             // ecx = count, eax = source
  jmp  item_move_force_
end:
  inc  eax
  retn
 }
}

static void __declspec(naked) move_table_target() {
 __asm {
  mov  edx, ds:[_target_curr_stack]
  mov  edx, ds:[_target_stack][edx*4]
  jmp  move_table_source
 }
}

static void __declspec(naked) proto_ptr_call() {
 __asm {
  call proto_ptr_
  mov  edx, [esp+0x4]
  mov  eax, [edx+0x70]                      // eax = p_item.size
  test eax, eax
  jnz  end
  cmp  dword ptr [edx+0x20], item_type_container
  jne  end
  mov  eax, [edx+0x24]                      // max_size
  shr  eax, 1                               // eax = max_size/2
  inc  eax
  mov  [edx+0x70], eax                      // container.size = max_size/2 + 1
end:
  retn
 }
}

void InventoryInit() {

 //  STAT_unused   InvSizeLimitMode  STAT_size   
 MakeCall(0x4AF0CB, &stat_level_hook, true);
 //     STAT_unused    
 MakeCall(0x455D65, &op_set_critter_stat_hook, false);
 //     STAT_unused
 MakeCall(0x4AF54E, &stat_set_base_hook, false);

 //     /  ,       
 MakeCall(0x4725C8, &display_stats_hook, true);

 InvSizeLimitMode = GetPrivateProfileInt("Misc", "CritterInvSizeLimitMode", 0, ini);
 if (InvSizeLimitMode >= 1 && InvSizeLimitMode <= 7) {
  if (InvSizeLimitMode >= 4) {
   InvSizeLimitMode -= 4;
   SafeWrite8(0x477EB3, 0xEB);
  }
  if (InvSizeLimitMode) {
   InvSizeLimit = GetPrivateProfileInt("Misc", "CritterInvSizeLimit", 50, ini);

   HookCall(0x42E67E, &critterIsOverloaded_hook);

   //Check capacity of player and barteree when bartering
   HookCall(0x474C73, &barter_attempt_transaction_peon);
   HookCall(0x474CCA, &barter_attempt_transaction_barterer);

   //  "Take all"
   HookCall(0x47410B, &loot_container_hook);

   //        "" 
   HookCall(0x449136, &gdControlUpdateInfo_hook);

   //Display item weight when examening
   GetPrivateProfileString("sfall", "ItemSizeMsg", "It occupies %d unit(s).", ItemSizeMsg, 32, translationIni);
   HookCall(0x473007, &inven_obj_examine_func_hook);
  }
 } else InvSizeLimitMode = 0;

 if (GetPrivateProfileInt("Misc", "SuperStimExploitFix", 0, ini)) {
  GetPrivateProfileString("sfall", "SuperStimExploitMsg", "You cannot use a super stim on someone who is not injured!", SuperStimMsg, 128, translationIni);
  MakeCall(0x49C3DE, &protinst_use_item_on_hook, false);
 }

 invenapcost = GetPrivateProfileInt("Misc", "InventoryApCost", 4, ini);
 invenapqpreduction = GetPrivateProfileInt("Misc", "QuickPocketsApCostReduction", 2, ini);
 MakeCall(0x46E80B, &handle_inventory_hook, true);

 if (GetPrivateProfileInt("Misc", "CheckWeaponAmmoCost", 0, ini)) HookCall(0x4266E9, &combat_check_bad_shot_hook);

 ReloadWeaponKey = GetPrivateProfileIntA("Input", "ReloadWeaponKey", 0, ini);
 if (ReloadWeaponKey) HookCall(0x442DFF, &ReloadWeaponHotKey);

 if (GetPrivateProfileInt("Misc", "AutoReloadWeapon", 0, ini)) HookCall(0x422E8A, &AutoReloadWeapon);

 HookCall(0x45419B, &correctFidForRemovedItem_hook);
 HookCall(0x4494FC, &ControlWeapon_hook);
 HookCall(0x449570, &ControlArmor_hook);

 if (GetPrivateProfileInt("Misc", "FreeWeight", 0, ini)) {
  MakeCall(0x47002D, &display_inventory_hook, true);
  MakeCall(0x4703E9, &display_target_inventory_hook, true);

  HookCall(0x475363, &display_table_inventories_hook);

  SafeWrite16(0x4753D5, 0xD231);
  MakeCall(0x47541F, &display_table_inventories_hook1, true);
  HookCall(0x47558A, &display_table_inventories_hook2);

  SafeWrite16(0x4755BF, 0xFF31);
  MakeCall(0x47560A, &display_table_inventories_hook3, true);
  HookCall(0x4757D2, &display_table_inventories_hook2);

  HookCall(0x475D93, &barter_inventory_hook);
 }

 if (GetPrivateProfileInt("Misc", "EquipArmor", 0, ini)) HookCall(0x4466CC, &gdProcess_hook);

 char buf[MAX_PATH-3];
 int i, count;
 memset(armors, 0, sizeof(npcArmor)*PartyMax);
 GetPrivateProfileString("Misc", "ArmorFile", "", buf, MAX_PATH, ini);
 if (strlen(buf) > 0) {
  sprintf(iniArmor, ".\\%s", buf);
  count = GetPrivateProfileIntA("Main", "Count", 0, iniArmor);
  char section[4];
  for (i = 1; i <= count; i++) {
   _itoa_s(i, section, 10);
   armors[npcCount].Pid = GetPrivateProfileIntA(section, "PID", 0, iniArmor);
   armors[npcCount].Default = GetPrivateProfileIntA(section, "Default", 0, iniArmor);
   armors[npcCount].Leather = GetPrivateProfileIntA(section, "Leather", 0, iniArmor);
   armors[npcCount].Power = GetPrivateProfileIntA(section, "Power", 0, iniArmor);
   armors[npcCount].Advanced = GetPrivateProfileIntA(section, "Advanced", 0, iniArmor);
   armors[npcCount].Metal = GetPrivateProfileIntA(section, "Metal", 0, iniArmor);
   armors[npcCount].Cured = GetPrivateProfileIntA(section, "Cured", 0, iniArmor);
   armors[npcCount].Combat = GetPrivateProfileIntA(section, "Combat", 0, iniArmor);
   armors[npcCount].Robe = GetPrivateProfileIntA(section, "Robe", 0, iniArmor);
   npcCount++;
  }
  if (npcCount > 0) HookCall(0x472833, &invenWieldFunc_hook);
 }

 StackEmptyWeapons = GetPrivateProfileIntA("Misc", "StackEmptyWeapons", 0, ini);
 if (StackEmptyWeapons) MakeCall(0x4736C6, &inven_action_cursor_hook, true);

//          
 int ReloadReserve = GetPrivateProfileIntA("Misc", "ReloadReserve", 1, ini);
 if (ReloadReserve >= 0) {
  SafeWrite32(0x47655F, ReloadReserve);      // mov  eax, ReloadReserve
  SafeWrite32(0x476563, 0x097EC139);         // cmp  ecx, eax; jle  0x476570
  SafeWrite16(0x476567, 0xC129);             // sub  ecx, eax
  SafeWrite8(0x476569, 0x91);                // xchg ecx, eax
 };

//       
 HookCall(0x471457, &inven_pickup_hook2);

//  " "
 MakeCall(0x46FA7A, &make_drop_button, false);
 MakeCall(0x473EFC, &drop_all, true);
 HookCall(0x47418D, &loot_container_hook1);
 GetPrivateProfileString("sfall", "OverloadedDrop", "Sorry, there is no space left.", OverloadedDrop, 48, translationIni);

 UseScrollWheel = GetPrivateProfileIntA("Input", "UseScrollWheel", 1, ini) != 0;
 if (UseScrollWheel) {
  MakeCall(0x473E66, &loot_container_hook2, false);
  MakeCall(0x4759F1, &barter_inventory_hook1, false);
  *((DWORD*)_max) = 100;
 };

//    (   )    
 HookCall(0x46EB1E, &handle_inventory_hook1);

//    
 MakeCall(0x477159, &item_add_mult, true);

//       ,    
 SafeWrite32(0x475CF2, _stack);

//            ,   
 HookCall(0x475281, &move_table_source);
 HookCall(0x4752F8, &move_table_target);

//        ,    
 HookCall(0x47531F, (void*)gdialogDisplayMsg_);
 SafeWrite16(0x475300, 0x87EB);             // jmps 0x475289

 if (GetPrivateProfileIntA("Misc", "ContainerSizeFix", 0, ini)) {
  HookCall(0x477B75, &proto_ptr_call);
  HookCall(0x479A4A, &proto_ptr_call);
 }

}

void InventoryReset() {
 invenapcost = GetPrivateProfileInt("Misc", "InventoryApCost", 4, ini);
}
