/*
 *    sfall
 *    Copyright (C) 2013  The sfall team
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.h"

#include <vector>
#include "Define.h"
#include "FalloutEngine.h"
#include "PartyControl.h"

DWORD IsControllingNPC = 0;
DWORD HiddenArmor = 0;

static DWORD Mode;
static std::vector<WORD> Chars;

static char real_pc_name[32];
static DWORD real_last_level;
static DWORD real_Level;
static DWORD real_Experience;
static DWORD real_free_perk;
static DWORD real_unspent_skill_points;
static DWORD real_map_elevation;
static DWORD real_sneak_working;
static DWORD real_sneak_queue_time;
static DWORD real_dude;
static DWORD real_hand;
static DWORD real_trait;
static DWORD real_trait2;
static DWORD real_itemButtonItems[(6*4)*2];
static DWORD real_perkLevelDataList[PERK_count];
static DWORD real_drug_gvar[6];
static DWORD real_jet_gvar;
static DWORD real_tag_skill[4];

static void __declspec(naked) isRealParty() {
 __asm {
  push edx
  push ecx
  push ebx
  mov  ecx, ds:[_partyMemberMaxCount]
  dec  ecx
  jecxz skip                                //  
  mov  edx, ds:[_partyMemberPidList]
  mov  ebx, [eax+0x64]                      // pid
loopPid:
  add  edx, 4
  cmp  ebx, [edx]
  je   end
  loop loopPid
skip:
  xor  eax, eax
end:
  pop  ebx
  pop  ecx
  pop  edx
  retn
 }
}

void __declspec(naked) PartyControl_CanUseWeapon() {
 __asm {
  push edi
  push esi
  push edx
  push ecx
  push ebx
  mov  edi, eax
  call item_get_type_
  cmp  eax, item_type_weapon
  jne  canUse                               // 
  dec  eax
  dec  eax
  cmp  IsControllingNPC, eax                //  ?
  jne  end                                  // 
  inc  eax                                  // hit_right_weapon_primary
  xchg ecx, eax
  mov  eax, edi                             // eax=item
  call item_w_anim_code_
  xchg ecx, eax                             // ecx=ID1=Weapon code
  xchg edx, eax
  xchg edi, eax                             // eax=item
  call item_w_anim_weap_
  xchg ebx, eax                             // ebx=ID2=Animation code
  mov  esi, ds:[_inven_dude]
  mov  edx, [esi+0x20]                      // fid
  and  edx, 0xFFF                           // edx=Index
  mov  eax, [esi+0x1C]                      // cur_rot
  inc  eax
  push eax                                  // ID3=Direction code
  mov  eax, ObjType_Critter
  call art_id_
  call art_exists_
  test eax, eax
  jz   end
canUse:
  xor  eax, eax
  inc  eax
end:
  pop  ebx
  pop  ecx
  pop  edx
  pop  esi
  pop  edi
  retn
 }
}

static bool _stdcall IsInPidList(DWORD* npc) {
 if (Chars.size() == 0) return true;
 int pid = npc[0x64/4] & 0xFFFFFF;
 for (std::vector<WORD>::iterator it = Chars.begin(); it != Chars.end(); it++) {
  if (*it == pid) return true;
 }
 return false;
}

// save "real" dude state
static void __declspec(naked) SaveDudeState() {
 __asm {
  push edi
  push esi
  push edx
  push ecx
  push ebx
  mov  esi, _pc_name
  mov  edi, offset real_pc_name
  mov  ecx, 32/4
  rep  movsd
  mov  eax, ebx
  call critter_name_
  call critter_pc_set_name_
  mov  eax, ds:[_last_level]
  mov  real_last_level, eax
  mov  eax, ds:[_Level_]
  mov  real_Level, eax
  mov  eax, ebx
  call isRealParty
  test eax, eax
  jz   saveLevel
  call partyMemberGetCurLevel_
saveLevel:
  mov  ds:[_Level_], eax
  mov  ds:[_last_level], eax
  mov  eax, ds:[_Experience_]
  mov  real_Experience, eax
  movzx eax, byte ptr ds:[_free_perk]
  mov  real_free_perk, eax
  mov  eax, ds:[_curr_pc_stat]
  mov  real_unspent_skill_points, eax
  mov  eax, ds:[_map_elevation]
  mov  real_map_elevation, eax
  mov  eax, ds:[_sneak_working]
  mov  real_sneak_working, eax
  mov  eax, ds:[_obj_dude]
  push eax
  mov  edx, 10
  call queue_find_first_
  test edx, edx
  jz   noSneak
  mov  edx, [edx]
  mov  eax, ds:[_fallout_game_time]
  sub  edx, eax
  pop  eax
  push eax
  push edx
  mov  edx, 10
  call queue_remove_this_
  pop  edx
noSneak:
  mov  real_sneak_queue_time, edx
  pop  eax
  mov  real_dude, eax
  mov  eax, ds:[_itemCurrentItem]
  mov  real_hand, eax
  mov  edx, offset real_trait2
  mov  eax, offset real_trait
  call trait_get_
  mov  esi, _itemButtonItems
  mov  edi, offset real_itemButtonItems
  mov  ecx, (6*4)*2
  rep  movsd
  mov  esi, ds:[_perkLevelDataList]
  push esi
  mov  edi, offset real_perkLevelDataList
  mov  ecx, PERK_count
  push ecx
  rep  movsd
  pop  ecx
  pop  edi
  mov  eax, ebx
  call isRealParty
  test eax, eax
  jz   clearPerks
  call perkGetLevelData_
  xchg esi, eax
  rep  movsd
  xor  eax, eax
  jmp  skipPerks
clearPerks:
  rep  stosd
skipPerks:
  mov  ds:[_free_perk], al
  mov  ds:[_curr_pc_stat], eax
  mov  ds:[_sneak_working], eax
  mov  eax, [ebx+0x64]
  mov  ds:[_inven_pid], eax
  mov  ds:[_obj_dude], ebx
  mov  ds:[_inven_dude], ebx
  mov  ds:[_stack], ebx
  mov  esi, ds:[_game_global_vars]
  add  esi, 21*4                            // esi->GVAR_NUKA_COLA_ADDICT
  push esi
  mov  edi, offset real_drug_gvar
  mov  ecx, 6
  push ecx
  rep  movsd
  pop  ecx
  pop  edi
  mov  eax, [edi+296*4-21*4]                // eax = GVAR_ADDICT_JET
  mov  real_jet_gvar, eax
  push edi
  mov  edx, _drugInfoList
  mov  esi, ebx                             // _obj_dude
loopDrug:
  mov  eax, [edx]                           // eax = drug_pid
  call item_d_check_addict_
  test eax, eax                             //  ?
  jz   noAddict                             // 
  xor  eax, eax
  inc  eax
noAddict:
  mov  [edi], eax
  add  edx, 12
  add  edi, 4
  loop loopDrug
  test eax, eax                             //     ()?
  jnz  skipBooze                            // 
  mov  eax, [edx]                           // PID_BOOZE
  call item_d_check_addict_
  mov  [edi-4], eax                         // GVAR_ALCOHOL_ADDICT
skipBooze:
  mov  eax, [edx+12]                        // PID_JET
  call item_d_check_addict_
  test eax, eax
  jz   noJetAddict
  xor  eax, eax
  inc  eax
noJetAddict:
  pop  edi
  mov  [edi+296*4-21*4], eax                // GVAR_ADDICT_JET
  xor  eax, eax
  dec  eax
  call item_d_check_addict_
  mov  edx, 4
  test eax, eax
  mov  eax, edx
  jz   unsetAddict
  call pc_flag_on_
  jmp  skip
unsetAddict:
  call pc_flag_off_
skip:
  push edx
  mov  eax, offset real_tag_skill
  call skill_get_tags_
  mov  edi, _tag_skill
  pop  ecx
  xor  eax, eax
  dec  eax
  rep  stosd
  mov  edx, eax
  call trait_set_
  mov  ds:[_Experience_], eax
// get active hand by weapon anim code
  mov  edx, [ebx+0x20]                      // fid
  and  edx, 0x0F000
  sar  edx, 0xC                             // edx = current weapon anim code as seen in hands
  xor  ecx, ecx                             //  
  mov  eax, ebx
  call inven_right_hand_
  test eax, eax                             //     ?
  jz   setActiveHand                        // 
  push eax
  call item_get_type_
  cmp  eax, item_type_weapon
  pop  eax
  jne  setActiveHand                        // 
  call item_w_anim_code_
  cmp  eax, edx                             //  ?
  jne  setActiveHand                        // 
  inc  ecx                                  //  
setActiveHand:
  mov  ds:[_itemCurrentItem], ecx
  mov  eax, ebx
  call inven_right_hand_
  test eax, eax                             //     ?
  jz   noRightHand                          // 
  push eax
  call PartyControl_CanUseWeapon
  test eax, eax
  pop  eax
  jnz  noRightHand
  and  byte ptr [eax+0x27], 0xFD            //      
noRightHand:
  xchg ebx, eax
  call inven_left_hand_
  test eax, eax                             //     ?
  jz   noLeftHand                           // 
  push eax
  call PartyControl_CanUseWeapon
  test eax, eax
  pop  eax
  jnz  noLeftHand
  and  byte ptr [eax+0x27], 0xFE            //      
noLeftHand:
  pop  ebx
  pop  ecx
  pop  edx
  pop  esi
  pop  edi
  retn
 }
}

// restore dude state
static void __declspec(naked) RestoreDudeState() {
 __asm {
  pushad
  mov  eax, offset real_pc_name
  call critter_pc_set_name_
  mov  eax, real_last_level
  mov  ds:[_last_level], eax
  mov  eax, real_Level
  mov  ds:[_Level_], eax
  mov  eax, real_Experience
  mov  ds:[_Experience_], eax
  mov  eax, real_free_perk
  mov  ds:[_free_perk], al
  mov  eax, real_unspent_skill_points
  mov  ds:[_curr_pc_stat], eax
  mov  eax, real_map_elevation
  mov  ds:[_map_elevation], eax
  mov  eax, real_sneak_working
  mov  ds:[_sneak_working], eax
  mov  eax, real_hand
  mov  ds:[_itemCurrentItem], eax
  mov  edx, real_trait2
  mov  eax, real_trait
  call trait_set_
  mov  esi, offset real_itemButtonItems
  mov  edi, _itemButtonItems
  mov  ecx, (6*4)*2
  rep  movsd
  mov  esi, offset real_drug_gvar
  mov  edi, ds:[_game_global_vars]
  mov  eax, real_jet_gvar
  mov  [edi+296*4], eax                     // GVAR_ADDICT_JET
  add  edi, 21*4                            // esi->GVAR_NUKA_COLA_ADDICT
  mov  ecx, 6
  rep  movsd
  mov  edx, 4
  mov  eax, offset real_tag_skill
  call skill_set_tags_
  mov  eax, ds:[_obj_dude]
  mov  ecx, real_dude
  push ecx
  mov  ds:[_obj_dude], ecx
  mov  ds:[_inven_dude], ecx
  mov  ds:[_stack], ecx
  mov  esi, [ecx+0x64]
  mov  ds:[_inven_pid], esi
  mov  ecx, PERK_count
  push ecx
  mov  esi, ds:[_perkLevelDataList]
  push esi
  call isRealParty
  test eax, eax
  jz   skipPerks
  call perkGetLevelData_
  xchg edi, eax
  rep  movsd
skipPerks:
  pop  edi
  pop  ecx
  mov  esi, offset real_perkLevelDataList
  rep  movsd
  xor  eax, eax
  mov  IsControllingNPC, eax
  dec  eax
  call item_d_check_addict_
  test eax, eax
  mov  eax, 4
  jz   unsetAddict
  call pc_flag_on_
  jmp  skip
unsetAddict:
  call pc_flag_off_
skip:
  pop  edx
  mov  eax, real_sneak_queue_time
  test eax, eax
  jz   end
  mov  ecx, 10
  xor  ebx, ebx
  call queue_add_
end:
  popad
  retn
 }
}

void __declspec(naked) PartyControl_PrintWarning() {
 __asm {
  push edx
  push ebx
  sub  esp, 16
  mov  edx, esp
  mov  ebx, 675                             // 'I can't do that.'/'    .'
  mov  eax, _proto_main_msg_file
  call getmsg_
  call display_print_
  xor  eax, eax
  add  esp, 16
  pop  ebx
  pop  edx
  retn
 }
}

static void _declspec(naked) CombatWrapper_v2() {
 __asm {
  pushad
  cmp  eax, ds:[_obj_dude]
  jne  skip
  xor  edx, edx
  cmp  ds:[_combatNumTurns], edx
  je   skipControl                          //   
  mov  eax, [eax+0x4]                       // tile_num
  add  edx, 2
  call tile_scroll_to_
  jmp  skipControl
skip:
  push eax
  call IsInPidList
  and  eax, 0xFF
  test eax, eax
  jz   skipControl
  cmp  Mode, eax                            // control all critters?
  je   npcControl
  popad
  pushad
  call isPartyMember_
  test eax, eax
  jnz  npcControl
skipControl:
  popad
  jmp  combat_turn_
npcControl:
  mov  IsControllingNPC, eax
  popad
  pushad
  xchg ebx, eax                             // ebx = npc
  call SaveDudeState
  call intface_redraw_
  mov  eax, [ebx+0x4]                       // tile_num
  mov  edx, 2
  call tile_scroll_to_
  xchg ebx, eax                             // eax = npc
  xor  edx, edx
  push edx
  call combat_turn_
  pop  ecx
  xchg ecx, eax
  cmp  IsControllingNPC, eax                // if game was loaded during turn, PartyControlReset()
  je   skipRestore                          // was called and already restored state
  call RestoreDudeState
  call intface_redraw_
skipRestore:
  test ecx, ecx                             //   ?
  popad
  jz   end                                  // 
// ///
  test byte ptr [eax+0x44], 0x80            // DAM_DEAD
  jnz  end
  xor  eax, eax
  dec  eax
  retn
end:
  xor  eax, eax
  retn
 }
}

// hack to exit from this function safely when you load game during NPC turn
static void _declspec(naked) combat_add_noncoms_hook() {
 __asm {
  call CombatWrapper_v2
  inc  eax
  jnz  end
  mov  ds:[_list_com], eax
  mov  ecx, [esp+0x4]
end:
  retn
 }
}

static void _declspec(naked) PrintLevelWin_hook() {
 __asm {
  xor  edi, edi
  cmp  IsControllingNPC, edi
  je   end
  dec  edi
  xchg edi, eax
end:
  mov  edi, eax
  cmp  eax, -1
  retn
 }
}

static void _declspec(naked) Save_as_ASCII_hook() {
 __asm {
  xor  ebx, ebx
  cmp  IsControllingNPC, ebx
  je   end
  dec  ebx
  xchg ebx, eax
end:
  mov  ebx, 649
  retn
 }
}

static void _declspec(naked) handle_inventory_hook() {
 __asm {
  xor  ebx, ebx
  mov  edx, eax                             // edx = _inven_dude
  call inven_worn_
  test eax, eax
  jz   end
  cmp  IsControllingNPC, ebx
  je   end
  push eax
  push edx
  inc  ebx
  xchg edx, eax                             // eax = source, edx = armor
  call item_remove_mult_
  pop  edx                                  // edx = source
  pop  ebx                                  // ebx = armor
  inc  eax                                  // ?
  jnz  nextArmor                            // 
//   ,       
  push edx
  push eax
  xchg ebx, eax                             // ebx = newarmor, eax = oldarmor
  xchg edx, eax                             // edx = oldarmor, eax = source
  call adjust_ac_
  pop  ebx
  pop  edx
nextArmor:
  mov  eax, edx
  call inven_worn_
  test eax, eax
  jz   noArmor
  and  byte ptr [eax+0x27], 0xFB            //    
  jmp  nextArmor
noArmor:
  xchg ebx, eax                             // eax = armor
  mov  HiddenArmor, eax
end:
  retn
 }
}

static void _declspec(naked) handle_inventory_hook1() {
 __asm {
  xor  edx, edx
  cmp  IsControllingNPC, edx
  je   end
  push eax
  mov  edx, HiddenArmor
  test edx, edx
  jz   skip
  or   byte ptr [edx+0x27], 4               //    
  xor  ebx, ebx
  inc  ebx
  call item_add_force_
  xor  edx, edx
skip:
  mov  HiddenArmor, edx
  pop  eax
end:
  jmp  inven_worn_
 }
}

static void _declspec(naked) combat_input_hook() {
 __asm {
  cmp  ebx, 0x20                            // Space ( )?
  jne  skip
space:
  pop  ebx                                  //   
  mov  ebx, 0x20                            // Space ( )
  push 0x4228A8
  retn
skip:
  cmp  dword ptr IsControllingNPC, 0
  je   end
  cmp  ebx, 0xD                             // Enter ( )?
  je   space                                // 
end:  
  retn
 }
}

static void __declspec(naked) action_skill_use_hook() {
 __asm {
  cmp  eax, SKILL_SNEAK
  jne  skip
  xor  eax, eax
  cmp  IsControllingNPC, eax
  je   end
  call PartyControl_PrintWarning
skip:
  pop  eax                                  //   
  xor  eax, eax
  dec  eax
end:
  retn
 }
}

static void __declspec(naked) action_use_skill_on_hook() {
 __asm {
  cmp  IsControllingNPC, eax
  je   end
  jmp  PartyControl_PrintWarning
end:
  jmp  pc_flag_toggle_
 }
}

void PartyControlInit() {
 Mode = GetPrivateProfileIntA("Misc", "ControlCombat", 0, ini);
 if (Mode == 1 || Mode == 2) {
  char pidbuf[512];
  pidbuf[511]=0;
  if (GetPrivateProfileStringA("Misc", "ControlCombatPIDList", "", pidbuf, 511, ini)) {
   char* ptr = pidbuf;
   char* comma;
   while (true) {
    comma = strchr(ptr, ',');
    if (!comma) 
     break;
    *comma = 0;
    if (strlen(ptr) > 0)
     Chars.push_back((WORD)strtoul(ptr, 0, 0));
    ptr = comma + 1;
   }
   if (strlen(ptr) > 0)
    Chars.push_back((WORD)strtoul(ptr, 0, 0));
  }
  dlog_f(" Mode %d, Chars read: %d.", DL_INIT, Mode, Chars.size());
  HookCall(0x422354, &combat_add_noncoms_hook);
  HookCall(0x422E20, &CombatWrapper_v2);
  MakeCall(0x434AAC, &PrintLevelWin_hook, false);
  MakeCall(0x43964E, &Save_as_ASCII_hook, false);
  HookCall(0x46E8BA, &handle_inventory_hook);
  HookCall(0x46EC0A, &handle_inventory_hook1);
  SafeWrite32(0x471E48, 152);               //   152,   80 
  MakeCall(0x422879, &combat_input_hook, false);
  MakeCall(0x4124E0, &action_skill_use_hook, false);
  HookCall(0x41279A, &action_use_skill_on_hook);
 } else dlog(" Disabled.", DL_INIT);
}

void __stdcall PartyControlReset() {
 __asm {
  cmp  dword ptr IsControllingNPC, 0
  je   end
  call RestoreDudeState
end:
 }
}
